import os
import yaml
import argparse
import numpy as np
import torch
from torchsummaryX import summary

from utils.helpers import create_logger, save_experiment
from utils.data_loading import import_data
from utils.model_loading import load_model
from utils.model_training import train_model
from utils.results_export import export_results


if __name__ == "__main__":

    # Load configuration
    parser = argparse.ArgumentParser(description="SIGCOMM_CCR")
    parser.add_argument(
        "-c", "--config", default="configuration/config.yml", help="Configuration File"
    )
    args = parser.parse_args()
    with open(args.config, "r") as config_file:
        configuration = yaml.safe_load(config_file)

    os.environ["CUDA_VISIBLE_DEVICES"] = configuration["gpu"]

    # Save experiment
    xp_dir = (
        "./results/"
        + str(configuration["dataset"])
        + "/"
        + str(configuration["model_name"])
        + "/"
        + str(configuration["experiment_run"])
        + "/"
    )
    save_experiment(xp_dir, args.config)
    log, logclose = create_logger(log_filename=os.path.join(xp_dir, "experiment.log"))

    # Load dataset
    X_train, y_train, X_validation, y_validation, X_test, y_test = import_data(
        configuration["dataset"], xp_dir, preprocess=configuration["preprocessing"]
    )

    # Build model
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model = load_model(
        configuration, size=[X_train.shape[1]], nbclass=len(np.unique(y_train))
    )
    if configuration["model_name"] != "dt" and configuration["model_name"] != "rf":
        print(summary(model, torch.zeros((1, 1, X_train.shape[1]))))
        model = model.to(device)

    # Train model
    train_model(
        configuration,
        model,
        X_train,
        y_train,
        X_validation,
        y_validation,
        X_test,
        y_test,
        xp_dir=xp_dir,
        log=log,
    )

    # Export results
    results = export_results(
        model,
        configuration,
        X_train,
        y_train,
        X_validation,
        y_validation,
        X_test,
        y_test,
        xp_dir,
    )
    print(results)
    logclose()
