import torch
from torch import nn


def conv3x3(in_channels, out_channels, stride=1):
    """
    Convolution 3x3
    
    Parameters
    ----------
    in_channels: integer
        Number of input channels

    out_channels: integer
        Number of output channels
        
    stride: integer
        Stride of the convolution
    
    Returns
    -------
    feature_maps: array
        Feature maps from the convolution
    """

    return nn.Conv1d(
        in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False
    )


class ResidualBlock(nn.Module):
    """Class used to create a residual block"""

    def __init__(self, in_channels, out_channels, stride=1, downsample=None):
        """
        Parameters
        ----------
        in_channels: integer
            Number of input channels
            
        out_channels: integer
            Number of output channels
        
        stride: integer
            Stride of the convolution
            
        downsample: string
            Layer added in case of stride!=1 or in_channels!=out_channels
        """
        super(ResidualBlock, self).__init__()
        self.conv1 = conv3x3(in_channels, out_channels, stride)
        self.bn1 = nn.BatchNorm1d(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(out_channels, out_channels)
        self.bn2 = nn.BatchNorm1d(out_channels)
        self.downsample = downsample

    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        if self.downsample:
            residual = self.downsample(x)
        out += residual
        out = self.relu(out)
        return out


class ResNet(nn.Module):
    """Class used to create a ResNet model"""

    def __init__(self, size, nbclass, block=ResidualBlock, layers=[2, 2], ini=True):
        """
        Parameters
        ----------
        size: array
            Dimensions of the input data
                                
        nbclass: integer
            Number of classes
        
        block: string
            Name of the block used to build the network
            
        layers: array
            Number of layers per block
        
        ini: boolean
            Initialize weights if True
        """
        super(ResNet, self).__init__()
        self.in_channels = 16
        self.conv = conv3x3(1, 16)
        self.bn = nn.BatchNorm1d(16)
        self.relu = nn.ReLU(inplace=True)
        self.layer1 = self.make_layer(block, 32, layers[0])
        self.layer2 = self.make_layer(block, 64, layers[1], 1)
        self.fc = nn.Linear(size[0] * 64, nbclass)

        if ini:
            self.initialize_weights()

    def make_layer(self, block, out_channels, blocks, stride=1):
        downsample = None
        if (stride != 1) or (self.in_channels != out_channels):
            downsample = nn.Sequential(
                conv3x3(self.in_channels, out_channels, stride=stride),
                nn.BatchNorm1d(out_channels),
            )
        layers = []
        layers.append(block(self.in_channels, out_channels, stride, downsample))
        self.in_channels = out_channels
        for i in range(1, blocks):
            layers.append(block(out_channels, out_channels))
        return nn.Sequential(*layers)

    def forward(self, x):
        out = self.conv(x)
        out = self.bn(out)
        out = self.layer1(out)
        out = self.layer2(out)
        out = out.view(out.size(0), -1)
        out = self.fc(out)
        return out, x

    def initialize_weights(self) -> None:
        for m in self.modules():
            if isinstance(m, nn.Conv1d):
                nn.init.kaiming_uniform_(m.weight, nonlinearity="relu")
            elif isinstance(m, nn.Linear):
                nn.init.kaiming_uniform_(m.weight, nonlinearity="relu")


def model_resnet(configuration, size, nbclass):
    """
    Build ResNet model
    
    Parameters
    ----------
    configuration: array
        Elements from the configuration file
        
    size: array
        Dimensions of the input data
        
    nbclass: integer
        Number of classes

    Returns
    -------
    model: model
        ResNet model
    """
    model = ResNet(size=size, nbclass=nbclass)

    return model
