import torch
from torch import nn


class VanillaCNN(nn.Module):
    """Class used to create a vanilla 1D-CNN model"""

    def __init__(self, size, nbclass, ini=True):
        """
        Parameters
        ----------            
        size: array
            Dimensions of the input data
                                
        nbclass: integer
            Number of classes
        
        init: boolean
            Initialize the weights if True
        """
        super(VanillaCNN, self).__init__()
        self.conv1 = nn.Sequential(
            nn.Conv1d(
                in_channels=1, out_channels=32, kernel_size=3, stride=1, padding=1
            ),
            nn.ReLU(),
        )
        self.conv2 = nn.Sequential(nn.Conv1d(32, 64, 3, 1, 1), nn.ReLU())

        self.conv3 = nn.Sequential(nn.Conv1d(64, 64, 3, 1, 1), nn.ReLU())

        self.out = nn.Linear(size[0] * 64, nbclass)

        if ini:
            self.initialize_weights()

    def forward(self, x):
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = x.view(x.size(0), -1)
        output = self.out(x)

        return output, x

    def initialize_weights(self) -> None:
        for m in self.modules():
            if isinstance(m, nn.Conv1d):
                nn.init.kaiming_uniform_(m.weight, nonlinearity="relu")
            elif isinstance(m, nn.Linear):
                nn.init.kaiming_uniform_(m.weight, nonlinearity="relu")


def model_vanilla_cnn(configuration, size, nbclass):
    """
    Build Vanilla 1D-CNN model
    
    Parameters
    ----------
    configuration: array
        Elements from the configuration file
        
    size: array
        Dimensions of the input data
        
    nbclass: integer
        Number of classes

    Returns
    -------
    model: model
        Vanilla 1D-CNN model
    """
    model = VanillaCNN(size=size, nbclass=nbclass)

    return model
