import numpy as np
import pandas as pd
import torch
from torchvision import datasets, transforms
from torch.utils.data import TensorDataset, DataLoader
from sklearn import preprocessing


def load_data(dataset, preprocess=True):
    """
    Import train, validation and test sets
    
    Parameters
    ----------
    dataset: string
        Name of the dataset
        
    preprocess: boolean
        Perform preprocessing if True
    
    Returns
    -------
    sets: array
        Train, validation and test sets
    """
    ### Train set ###
    X_train = np.load("./data/" + dataset + "/train_x.npy")
    if preprocess:
        scaler1 = preprocessing.MinMaxScaler()
        X_train = scaler1.fit_transform(X_train)

    y_train = np.load("./data/" + dataset + "/train_y.npy")
    le = preprocessing.LabelEncoder()
    y_train = le.fit_transform(y_train)

    ### Validation set ###
    X_validation = np.load("./data/" + dataset + "/valid_x.npy")
    if preprocess:
        X_validation = scaler1.transform(X_validation)

    y_validation = np.load("./data/" + dataset + "/valid_y.npy")
    y_validation = le.transform(y_validation)

    ### Test set ###
    X_test = np.load("./data/" + dataset + "/test_x.npy")
    if preprocess:
        X_test = scaler1.transform(X_test)

    y_test = np.load("./data/" + dataset + "/test_y.npy")
    y_test = le.transform(y_test)

    return X_train, y_train, X_validation, y_validation, X_test, y_test


def import_data(dataset, xp_dir, preprocess=True):
    """
    Generate train, validation and test sets
    
    Parameters
    ----------
   dataset: string
        Name of the dataset
        
    xp_dir: string
        Folder of the experiment
        
    preprocess: boolean
        Perform preprocessing if True
    
    Returns
    -------
    sets: array
        Train, validation and test sets
    """
    X_train, y_train, X_validation, y_validation, X_test, y_test = load_data(
        dataset, preprocess=preprocess
    )

    print("Train set size: ", X_train.shape)
    print("Validation set size: ", X_validation.shape)
    print("Test set size: ", X_test.shape)

    return X_train, y_train, X_validation, y_validation, X_test, y_test


def get_tabular_data(X):
    """
    Generate tabular data
    
    Parameters
    ----------
    X: array
        Set
    
    Returns
    -------
    X_tabular: array
        Array in tabular format
    """
    X_tabular = np.squeeze(X)

    return X_tabular


def get_loaders(
    X_train,
    y_train,
    X_validation,
    y_validation,
    X_test,
    y_test,
    batch_size,
    shuffle=True,
    log=print,
):
    """
    Format arrays to data loaders
    
    Parameters
    ----------
    X_train: array
        Train set
        
    y_train: array
        Labels of the train set
        
    X_validation: array
        Validation set
        
    y_validation: array
        Labels of the validation set
        
    X_test: array
        Test set
        
    y_test: array
        Labels of the test set
        
    batch_size: integer
        Batch size
        
    Shuffle: boolean
        Shuffle the data if True
        
    log: string
        Processing of the outputs
    
    Returns
    -------
    loaders: tensor
        Train, validation and test loaders
    """
    # Parameters
    params = {"batch_size": batch_size, "shuffle": shuffle, "num_workers": 2}

    # Loaders
    train_set = TensorDataset(torch.Tensor(X_train), torch.Tensor(y_train))
    train_loader = DataLoader(train_set, **params)

    validation_set = TensorDataset(
        torch.Tensor(X_validation), torch.Tensor(y_validation)
    )
    validation_loader = DataLoader(validation_set, **params)

    test_set = TensorDataset(torch.Tensor(X_test), torch.Tensor(y_test))
    test_loader = torch.utils.data.DataLoader(test_set, **params)

    log("number of classes: {0}".format(len(np.unique(y_train))))
    log("batch size: {0}".format(batch_size))

    return train_loader, validation_loader, test_loader
