import os
import torch
from sklearn.metrics import accuracy_score

from utils.data_loading import get_loaders, get_tabular_data
import utils.train_and_test as tnt


def train_traditional_model(configuration, model, X_train, y_train, xp_dir, log=print):
    """
    Train traditional machine learning model
    
    Parameters
    ----------
    configuration: array
        Elements from the configuration file
    
    model: model
        Model to train
    
    X_train: array
        Train set
        
    y_train: array
        Labels of the train set
                
    xp_dir: string
        Folder of the experiment
        
    log: string
        Processing of the output
    """
    model.fit(X_train, y_train)


def train_dl_model(
    configuration,
    model,
    loader_train,
    loader_validation,
    loader_test,
    xp_dir,
    log=print,
):
    """
    Train deep learning model
    
    Parameters
    ----------
    configuration: array
        Elements from the configuration file
    
    model: model
        Model to train
    
    loader_train: tensor
        Loader of the train set
        
    loader_validation: tensor
        Loader of the validation set
        
    loader_test: tensor
        Loader of the test set
        
    xp_dir: string
        Folder of the experiment
        
    log: string
        Processing of the output    
    """
    optimizer = torch.optim.Adam(model.parameters(), lr=configuration["lr"])
    model_multi = torch.nn.DataParallel(model)

    for epoch in range(1, configuration["epochs"] + 1):
        log("Epoch: \t{0}".format(epoch))
        accu_train, accu_validation = tnt.train(
            model=model_multi,
            loader_train=loader_train,
            loader_validation=loader_validation,
            optimizer=optimizer,
            epoch=epoch,
            xp_dir=xp_dir,
            log=log,
        )
        accu_test, _ = tnt.test(
            model=model_multi,
            loader_test=loader_test,
            epoch=epoch,
            xp_dir=xp_dir,
            log=log,
        )


def train_model(
    configuration,
    model,
    X_train,
    y_train,
    X_validation,
    y_validation,
    X_test,
    y_test,
    xp_dir,
    log=print,
):
    """
    Train specified model
    
    Parameters
    ----------
    configuration: array
        Elements from the configuration file
    
    model: model
        Model to train
    
    X_train: array
        Train set
        
    y_train: array
        Labels of the train set
        
    X_validation: array
        Validation set
        
    y_validation: array
        Labels of the validation set
        
    X_test: array
        Test set
        
    y_test: array
        Labels of the test set
        
    xp_dir: string
        Folder of the experiment
        
    log: string
        Processing of the output
    """
    log("Start Training")

    if configuration["model_name"] != "dt" and configuration["model_name"] != "rf":
        loader_train, loader_validation, loader_test = get_loaders(
            X_train,
            y_train,
            X_validation,
            y_validation,
            X_test,
            y_test,
            batch_size=configuration["batch_size"],
            shuffle=True,
            log=log,
        )
        train_dl_model(
            configuration,
            model,
            loader_train,
            loader_validation,
            loader_test,
            xp_dir,
            log=log,
        )
    else:
        X_train_tabular = get_tabular_data(X_train)
        train_traditional_model(
            configuration, model, X_train_tabular, y_train, xp_dir, log=print
        )
