import torch
import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score

from utils.data_loading import get_loaders, get_tabular_data


def get_accuracies_dl(xp_dir):
    """
    Get accuracies for train/validation/test sets
    
    Parameters
    ----------
    xp_dir: string
        Folder of the experiment
    
    Returns
    -------
    accuracies: float
        Calculated accuracies
    """
    df = pd.read_table(xp_dir + "Accuracy.log", delimiter=" ", header=None)
    df.columns = ["Set", "Epoch", "Accuracy", "Accuracy_Validation"]
    df = df.replace("None", 0)
    df["Accuracy"] = df["Accuracy"].astype(float)
    df["Accuracy_Validation"] = df["Accuracy_Validation"].astype(float)

    epoch = df[df.Accuracy_Validation == df.Accuracy_Validation.max()].Epoch.values[0]
    df_best = df[df.Epoch == epoch].reset_index(drop=True)

    acc_train = round(df_best.loc[0, "Accuracy"] * 100, 1)
    acc_validation = round(df_best.loc[0, "Accuracy_Validation"] * 100, 1)
    acc_test = round(df_best.loc[1, "Accuracy"] * 100, 1)

    return acc_train, acc_validation, acc_test


def get_predictions(model, dataloader):
    """
    Get predictions 
    
    Parameters
    ----------
    model: model
        Model to process
        
    dataloader: tensor
        Data loader
    
    Returns
    -------
    sets: array
        Predictions along with labels 
    """
    model.eval()

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    y_true = torch.tensor([], dtype=torch.long, device=device)
    all_outputs = torch.tensor([], device=device)

    for i, (sample, label) in enumerate(dataloader):
        sample_input = sample.cuda()
        target = label.cuda()

        with torch.no_grad():
            output, _ = model(sample_input)
            y_true = torch.cat((y_true, target), 0)
            all_outputs = torch.cat((all_outputs, output), 0)

    y_true = y_true.cpu().numpy()
    _, y_pred = torch.max(all_outputs, 1)
    y_pred = y_pred.cpu().numpy()

    return y_true, y_pred


def export_results(
    model,
    configuration,
    X_train,
    y_train,
    X_validation,
    y_validation,
    X_test,
    y_test,
    xp_dir,
):
    """
    Export the results according to the model type
    
    Parameters
    ----------
    model: model
        Model to process
        
    configuration: array
        Elements from the configuration file
        
    X_train: array
        Train set
        
    y_train: array
        Labels of the train set
        
    X_validation: array
        Validation set
        
    y_validation: array
        Labels of the validation set
        
    X_test: array
        Test set
        
    y_test: array
        Labels of the test set
        
    xp_dir: string
        Folder of the experiment
    
    Returns
    -------
    results: array
        Results from the experiments
    """
    if configuration["model_name"] != "dt" and configuration["model_name"] != "rf":
        results = export_results_dl(
            model,
            configuration,
            X_train,
            y_train,
            X_validation,
            y_validation,
            X_test,
            y_test,
            xp_dir,
        )
    else:
        results = export_results_traditional(
            model,
            configuration,
            X_train,
            y_train,
            X_validation,
            y_validation,
            X_test,
            y_test,
            xp_dir,
        )

    return results


def export_results_traditional(
    model,
    configuration,
    X_train,
    y_train,
    X_validation,
    y_validation,
    X_test,
    y_test,
    xp_dir,
):
    """
    Export the results for a traditional machine learning model
    
    Parameters
    ----------
    model: model
        Model to process
        
    configuration: array
        Elements from the configuration file
        
    X_train: array
        Train set
        
    y_train: array
        Labels of the train set
        
    X_validation: array
        Validation set
        
    y_validation: array
        Labels of the validation set
        
    X_test: array
        Test set
        
    y_test: array
        Labels of the test set
        
    xp_dir: string
        Folder of the experiment
    
    Returns
    -------
    results: array
        Results from the experiments
    """
    # Accuracy
    acc_train = accuracy_score(y_train, model.predict(get_tabular_data(X_train)))
    acc_validation = accuracy_score(
        y_validation, model.predict(get_tabular_data(X_validation))
    )
    acc_test = accuracy_score(y_test, model.predict(get_tabular_data(X_test)))

    if configuration["model_name"] == "dt":
        depth = model.get_depth()
        leaves = model.get_n_leaves()
    if configuration["model_name"] == "rf":
        max_depth = list()
        for tree in model.estimators_:
            max_depth.append(tree.tree_.max_depth)
        depth = sum(max_depth) / len(max_depth)
        leaves = 0

    # Export results
    results_export = pd.DataFrame()
    results_export = results_export.append(
        [
            [
                configuration["dataset"],
                configuration["model_name"],
                X_train.shape,
                depth,
                leaves,
                acc_train,
                acc_validation,
                acc_test,
            ]
        ]
    )
    results_export.columns = [
        "Dataset",
        "Architecture",
        "Train_Size",
        "Depth",
        "Leaves",
        "Accuracy_Train",
        "Accuracy_Validation",
        "Accuracy_Test",
    ]
    results_export.to_csv(
        xp_dir
        + "/results_"
        + str(configuration["dataset"].split("/")[0])
        + "_"
        + str(configuration["dataset"].split("/")[1])
        + "_"
        + str(configuration["model_name"])
        + "_"
        + str(configuration["experiment_run"])
        + ".csv",
        index=False,
    )

    return results_export


def export_results_dl(
    model,
    configuration,
    X_train,
    y_train,
    X_validation,
    y_validation,
    X_test,
    y_test,
    xp_dir,
):
    """
    Export the results for a deep learning model
    
    Parameters
    ----------
    model: model
        Model to process
        
    configuration: array
        Elements from the configuration file
        
    X_train: array
        Train set
        
    y_train: array
        Labels of the train set
        
    X_validation: array
        Validation set
        
    y_validation: array
        Labels of the validation set
        
    X_test: array
        Test set
        
    y_test: array
        Labels of the test set
        
    xp_dir: string
        Folder of the experiment
    
    Returns
    -------
    results: array
        Results from the experiments
    """
    # Accuracy
    acc_train, acc_validation, acc_test = get_accuracies_dl(xp_dir)

    # Export results
    results_export = pd.DataFrame()
    results_export = results_export.append(
        [
            [
                configuration["dataset"],
                configuration["model_name"],
                configuration["preprocessing"],
                X_train.shape,
                configuration["epochs"],
                configuration["lr"],
                acc_train,
                acc_validation,
                acc_test,
            ]
        ]
    )
    results_export.columns = [
        "Dataset",
        "Architecture",
        "Preprocessing",
        "Train_Size",
        "Epochs",
        "LR",
        "Accuracy_Train",
        "Accuracy_Validation",
        "Accuracy_Test",
    ]
    results_export.to_csv(
        xp_dir
        + "/results_"
        + str(configuration["dataset"].split("/")[0])
        + "_"
        + str(configuration["dataset"].split("/")[1])
        + "_"
        + str(configuration["model_name"])
        + "_"
        + str(configuration["experiment_run"])
        + ".csv",
        index=False,
    )

    return results_export
