import time
import torch
import numpy as np


def train_or_test(
    model,
    dataloader,
    tot_set,
    epoch,
    xp_dir,
    loader_validation=None,
    optimizer=None,
    log=print,
):
    """
    Train or test the model
    
    Parameters
    ----------
    model: model
        Model to process
        
    dataloader: tensor
        Data loader
        
    tot_set: string
        Indicate train or test
        
    epoch: integer
        Number of epochs
                
    xp_dir: string
        Folder of the experiment
        
    loader_validation: tensor
        Loader of the validation set
        
    optimizer: array
        Optimizer setting
        
    log: string
        Processing of the output
    
    Returns
    -------
    accuracy: float
        Calculated accuracy
    """
    is_train = optimizer is not None
    start = time.time()
    n_examples = 0
    n_correct = 0
    n_examples_validation = 0
    n_correct_validation = 0
    n_batches = 0
    total_cross_entropy = 0
    min_valid_loss = np.inf
    accu_validation = None

    # Train / test
    for i, (sample, label) in enumerate(dataloader):
        sample_input = sample.cuda()
        sample_input = torch.unsqueeze(sample_input, 1)
        target = label.cuda()
        target = target.to(dtype=torch.long)

        grad_req = torch.enable_grad() if is_train else torch.no_grad()
        with grad_req:
            output, _ = model(sample_input)
            cross_entropy = torch.nn.functional.cross_entropy(output, target)
            _, predicted = torch.max(output.data, 1)
            n_examples += target.size(0)
            n_correct += (predicted == target).sum().item()
            n_batches += 1
            total_cross_entropy += cross_entropy.item()

        if is_train:
            loss = cross_entropy
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

        del sample_input
        del target
        del output
        del predicted

    # Validation
    if is_train:
        valid_loss = 0.0
        model.eval()

        for i, (sample, label) in enumerate(loader_validation):
            sample_input = sample.cuda()
            sample_input = torch.unsqueeze(sample_input, 1)
            target = label.cuda()
            target = target.to(dtype=torch.long)

            grad_req = torch.enable_grad() if is_train else torch.no_grad()
            with grad_req:
                output, _ = model(sample_input)
                cross_entropy = torch.nn.functional.cross_entropy(output, target)
                _, predicted = torch.max(output.data, 1)
                n_examples_validation += target.size(0)
                n_correct_validation += (predicted == target).sum().item()
                valid_loss += cross_entropy.item()
        if min_valid_loss > valid_loss:
            min_valid_loss = valid_loss

    end = time.time()

    log("\tTime: \t{0}".format(end - start))
    log("\tCross entropy: \t{0}".format(total_cross_entropy / n_batches))
    log("\tAccuracy: \t\t{0}%".format(n_correct / n_examples * 100))
    if is_train:
        log(
            "\tAccuracy validation: \t\t{0}%".format(
                n_correct_validation / n_examples_validation * 100
            )
        )
        accu_validation = n_correct_validation / n_examples_validation

    f = open(xp_dir + "Accuracy.log", "a")
    text = (
        tot_set
        + " "
        + str(epoch)
        + " "
        + str(n_correct / n_examples)
        + " "
        + str(accu_validation)
    )
    f.write(text + "\n")
    f.close

    return n_correct / n_examples, accu_validation


def train(model, loader_train, epoch, xp_dir, loader_validation, optimizer, log=print):
    """
    Launch train
    
    Parameters
    ----------
    model: model
        Model to train
        
    loader_train: tensor
        Loader of the train set
     
    epoch: integer
        Number of epochs   
             
    xp_dir: string
        Folder of the experiment
    
    loader_validation: tensor
        Loader of the validation set
        
    optimizer: array
        Optimizer setting
            
    log: string
        Processing of the output
    
    Returns
    -------
    accuracy: float
        Train accuracy
    """
    log("\ttrain")
    model.train()

    return train_or_test(
        model=model,
        dataloader=loader_train,
        tot_set="Train",
        epoch=epoch,
        xp_dir=xp_dir,
        loader_validation=loader_validation,
        optimizer=optimizer,
        log=log,
    )


def test(model, loader_test, epoch, xp_dir, log=print):
    """
    Launch test
    
    Parameters
    ----------
    model: model
        Model to test
        
    loader_test: tensor
        Loader of the test set

    epoch: integer
        Number of epochs
        
    xp_dir: string
        Folder of the experiment
            
    log: string
        Processing of the output
    
    Returns
    -------
    accuracy: float
        Test accuracy
    """
    log("\ttest")
    model.eval()

    return train_or_test(
        model=model,
        dataloader=loader_test,
        tot_set="Test",
        epoch=epoch,
        xp_dir=xp_dir,
        optimizer=None,
        log=log,
    )
